import pytest

from sanic.log import deprecation


def test_deprecation():
    message = r"\[DEPRECATION v9\.9\] hello"
    with pytest.warns(DeprecationWarning, match=message):
        deprecation("hello", 9.9)



import logging

from unittest.mock import Mock

import pytest

from sanic import Sanic
from sanic.response import text
from sanic.server.async_server import AsyncioServer
from sanic.signals import Event
from sanic.touchup.schemes.ode import OptionalDispatchEvent


try:
    from unittest.mock import AsyncMock
except ImportError:
    from tests.asyncmock import AsyncMock  # type: ignore


@pytest.fixture
def app_one():
    app = Sanic("One")

    @app.get("/one")
    async def one(request):
        return text("one")

    return app


@pytest.fixture
def app_two():
    app = Sanic("Two")

    @app.get("/two")
    async def two(request):
        return text("two")

    return app


@pytest.fixture(autouse=True)
def clean():
    Sanic._app_registry = {}
    yield


def test_serve_same_app_multiple_tuples(app_one, run_multi):
    app_one.prepare(port=23456)
    app_one.prepare(port=23457)

    logs = run_multi(app_one)
    assert (
        "sanic.root",
        logging.INFO,
        "Goin' Fast @ http://127.0.0.1:23456",
    ) in logs
    assert (
        "sanic.root",
        logging.INFO,
        "Goin' Fast @ http://127.0.0.1:23457",
    ) in logs


def test_serve_multiple_apps(app_one, app_two, run_multi):
    app_one.prepare(port=23456)
    app_two.prepare(port=23457)

    logs = run_multi(app_one)
    assert (
        "sanic.root",
        logging.INFO,
        "Goin' Fast @ http://127.0.0.1:23456",
    ) in logs
    assert (
        "sanic.root",
        logging.INFO,
        "Goin' Fast @ http://127.0.0.1:23457",
    ) in logs


def test_listeners_on_secondary_app(app_one, app_two, run_multi):
    app_one.prepare(port=23456)
    app_two.prepare(port=23457)

    before_start = AsyncMock()
    after_start = AsyncMock()
    before_stop = AsyncMock()
    after_stop = AsyncMock()

    app_two.before_server_start(before_start)
    app_two.after_server_start(after_start)
    app_two.before_server_stop(before_stop)
    app_two.after_server_stop(after_stop)

    run_multi(app_one)

    before_start.assert_awaited_once()
    after_start.assert_awaited_once()
    before_stop.assert_awaited_once()
    after_stop.assert_awaited_once()


@pytest.mark.parametrize(
    "events",
    (
        (Event.HTTP_LIFECYCLE_BEGIN,),
        (Event.HTTP_LIFECYCLE_BEGIN, Event.HTTP_LIFECYCLE_COMPLETE),
        (
            Event.HTTP_LIFECYCLE_BEGIN,
            Event.HTTP_LIFECYCLE_COMPLETE,
            Event.HTTP_LIFECYCLE_REQUEST,
        ),
    ),
)
def test_signal_synchronization(app_one, app_two, run_multi, events):
    app_one.prepare(port=23456)
    app_two.prepare(port=23457)

    for event in events:
        app_one.signal(event)(AsyncMock())

    run_multi(app_one)

    assert len(app_two.signal_router.routes) == len(events) + 1

    signal_handlers = {
        signal.handler
        for signal in app_two.signal_router.routes
        if signal.name.startswith("http")
    }

    assert len(signal_handlers) == 1
    assert list(signal_handlers)[0] is OptionalDispatchEvent.noop


def test_warning_main_process_listeners_on_secondary(
    app_one, app_two, run_multi
):
    app_two.main_process_start(AsyncMock())
    app_two.main_process_stop(AsyncMock())
    app_one.prepare(port=23456)
    app_two.prepare(port=23457)

    log = run_multi(app_one)

    message = (
        f"Sanic found 2 listener(s) on "
        "secondary applications attached to the main "
        "process. These will be ignored since main "
        "process listeners can only be attached to your "
        "primary application: "
        f"{repr(app_one)}"
    )

    assert ("sanic.error", logging.WARNING, message) in log


def test_no_applications():
    Sanic._app_registry = {}
    message = "Did not find any applications."
    with pytest.raises(RuntimeError, match=message):
        Sanic.serve()


def test_oserror_warning(app_one, app_two, run_multi, capfd):
    orig = AsyncioServer.__await__
    AsyncioServer.__await__ = Mock(side_effect=OSError("foo"))
    app_one.prepare(port=23456, workers=2)
    app_two.prepare(port=23457, workers=2)

    run_multi(app_one)

    captured = capfd.readouterr()
    assert (
        "An OSError was detected on startup. The encountered error was: foo"
    ) in captured.err

    AsyncioServer.__await__ = orig


def test_running_multiple_offset_warning(app_one, app_two, run_multi, capfd):
    app_one.prepare(port=23456, workers=2)
    app_two.prepare(port=23457)

    run_multi(app_one)

    captured = capfd.readouterr()
    assert (
        f"The primary application {repr(app_one)} is running "
        "with 2 worker(s). All "
        "application instances will run with the same number. "
        f"You requested {repr(app_two)} to run with "
        "1 worker(s), which will be ignored "
        "in favor of the primary application."
    ) in captured.err


def test_running_multiple_secondary(app_one, app_two, run_multi, capfd):
    app_one.prepare(port=23456, workers=2)
    app_two.prepare(port=23457)

    before_start = AsyncMock()
    app_two.before_server_start(before_start)
    run_multi(app_one)

    before_start.await_count == 2



import asyncio
import logging
import time

from multiprocessing import Process

import httpx


PORT = 42101


def test_no_exceptions_when_cancel_pending_request(app, caplog):
    app.config.GRACEFUL_SHUTDOWN_TIMEOUT = 1

    @app.get("/")
    async def handler(request):
        await asyncio.sleep(5)

    @app.after_server_start
    def shutdown(app, _):
        time.sleep(0.2)
        app.stop()

    def ping():
        time.sleep(0.1)
        response = httpx.get("http://127.0.0.1:8000")
        print(response.status_code)

    p = Process(target=ping)
    p.start()

    with caplog.at_level(logging.INFO):
        app.run()

    p.kill()

    info = 0
    for record in caplog.record_tuples:
        assert record[1] != logging.ERROR
        if record[1] == logging.INFO:
            info += 1
        if record[2].startswith("Request:"):
            assert record[2] == (
                "Request: GET http://127.0.0.1:8000/ stopped. "
                "Transport is closed."
            )
    assert info == 11



import json
import subprocess

from pathlib import Path
from typing import List, Optional, Tuple

import pytest

from sanic_routing import __version__ as __routing_version__

from sanic import __version__


def capture(command: List[str]):
    proc = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=Path(__file__).parent,
    )
    try:
        out, err = proc.communicate(timeout=10)
    except subprocess.TimeoutExpired:
        proc.kill()
        out, err = proc.communicate()
    return out, err, proc.returncode


def starting_line(lines: List[str]):
    for idx, line in enumerate(lines):
        if line.strip().startswith(b"Sanic v"):
            return idx
    return 0


def read_app_info(lines: List[str]):
    for line in lines:
        if line.startswith(b"{") and line.endswith(b"}"):
            return json.loads(line)


@pytest.mark.parametrize(
    "appname,extra",
    (
        ("fake.server.app", None),
        ("fake.server:create_app", "--factory"),
        ("fake.server.create_app()", None),
    ),
)
def test_server_run(appname: str, extra: Optional[str]):
    command = ["sanic", appname]
    if extra:
        command.append(extra)
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    firstline = lines[starting_line(lines) + 1]

    assert exitcode != 1
    assert firstline == b"Goin' Fast @ http://127.0.0.1:8000"


def test_server_run_factory_with_args():
    command = [
        "sanic",
        "fake.server.create_app_with_args",
        "--factory",
    ]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")

    assert exitcode != 1, lines
    assert b"module=fake.server.create_app_with_args" in lines


def test_server_run_factory_with_args_arbitrary():
    command = [
        "sanic",
        "fake.server.create_app_with_args",
        "--factory",
        "--foo=bar",
    ]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")

    assert exitcode != 1, lines
    assert b"foo=bar" in lines


def test_error_with_function_as_instance_without_factory_arg():
    command = ["sanic", "fake.server.create_app"]
    out, err, exitcode = capture(command)
    assert b"try: \nsanic fake.server.create_app --factory" in err
    assert exitcode != 1


def test_error_with_path_as_instance_without_simple_arg():
    command = ["sanic", "./fake/"]
    out, err, exitcode = capture(command)
    assert (
        b"Please use --simple if you are passing a directory to sanic." in err
    )
    assert exitcode != 1


@pytest.mark.parametrize(
    "cmd",
    (
        (
            "--cert=certs/sanic.example/fullchain.pem",
            "--key=certs/sanic.example/privkey.pem",
        ),
        (
            "--tls=certs/sanic.example/",
            "--tls=certs/localhost/",
        ),
        (
            "--tls=certs/sanic.example/",
            "--tls=certs/localhost/",
            "--tls-strict-host",
        ),
    ),
)
def test_tls_options(cmd: Tuple[str]):
    command = ["sanic", "fake.server.app", *cmd, "-p=9999", "--debug"]
    out, err, exitcode = capture(command)
    assert exitcode != 1
    lines = out.split(b"\n")
    firstline = lines[starting_line(lines) + 1]
    assert firstline == b"Goin' Fast @ https://127.0.0.1:9999"


@pytest.mark.parametrize(
    "cmd",
    (
        ("--cert=certs/sanic.example/fullchain.pem",),
        (
            "--cert=certs/sanic.example/fullchain.pem",
            "--key=certs/sanic.example/privkey.pem",
            "--tls=certs/localhost/",
        ),
        ("--tls-strict-host",),
    ),
)
def test_tls_wrong_options(cmd: Tuple[str]):
    command = ["sanic", "fake.server.app", *cmd, "-p=9999", "--debug"]
    out, err, exitcode = capture(command)
    assert exitcode == 1
    assert not out
    lines = err.decode().split("\n")

    assert "TLS certificates must be specified by either of:" in lines


@pytest.mark.parametrize(
    "cmd",
    (
        ("--host=localhost", "--port=9999"),
        ("-H", "localhost", "-p", "9999"),
    ),
)
def test_host_port_localhost(cmd: Tuple[str]):
    command = ["sanic", "fake.server.app", *cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    expected = b"Goin' Fast @ http://localhost:9999"

    assert exitcode != 1
    assert expected in lines, f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize(
    "cmd",
    (
        ("--host=127.0.0.127", "--port=9999"),
        ("-H", "127.0.0.127", "-p", "9999"),
    ),
)
def test_host_port_ipv4(cmd: Tuple[str]):
    command = ["sanic", "fake.server.app", *cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    expected = b"Goin' Fast @ http://127.0.0.127:9999"

    assert exitcode != 1
    assert expected in lines, f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize(
    "cmd",
    (
        ("--host=::", "--port=9999"),
        ("-H", "::", "-p", "9999"),
    ),
)
def test_host_port_ipv6_any(cmd: Tuple[str]):
    command = ["sanic", "fake.server.app", *cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    expected = b"Goin' Fast @ http://[::]:9999"

    assert exitcode != 1
    assert expected in lines, f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize(
    "cmd",
    (
        ("--host=::1", "--port=9999"),
        ("-H", "::1", "-p", "9999"),
    ),
)
def test_host_port_ipv6_loopback(cmd: Tuple[str]):
    command = ["sanic", "fake.server.app", *cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    expected = b"Goin' Fast @ http://[::1]:9999"

    assert exitcode != 1
    assert expected in lines, f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize(
    "num,cmd",
    (
        (1, (f"--workers={1}",)),
        (2, (f"--workers={2}",)),
        (4, (f"--workers={4}",)),
        (1, ("-w", "1")),
        (2, ("-w", "2")),
        (4, ("-w", "4")),
    ),
)
def test_num_workers(num: int, cmd: Tuple[str]):
    command = ["sanic", "fake.server.app", *cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")

    if num == 1:
        expected = b"mode: production, single worker"
    else:
        expected = (f"mode: production, w/ {num} workers").encode()

    assert exitcode != 1
    assert expected in lines, f"Expected {expected}\nLines found: {lines}"


@pytest.mark.parametrize("cmd", ("--debug",))
def test_debug(cmd: str):
    command = ["sanic", "fake.server.app", cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    info = read_app_info(lines)

    assert info["debug"] is True, f"Lines found: {lines}\nErr output: {err}"
    assert (
        info["auto_reload"] is False
    ), f"Lines found: {lines}\nErr output: {err}"
    assert "dev" not in info, f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize("cmd", ("--dev", "-d"))
def test_dev(cmd: str):
    command = ["sanic", "fake.server.app", cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    info = read_app_info(lines)

    assert info["debug"] is True, f"Lines found: {lines}\nErr output: {err}"
    assert (
        info["auto_reload"] is True
    ), f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize("cmd", ("--auto-reload", "-r"))
def test_auto_reload(cmd: str):
    command = ["sanic", "fake.server.app", cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    info = read_app_info(lines)

    assert info["debug"] is False, f"Lines found: {lines}\nErr output: {err}"
    assert (
        info["auto_reload"] is True
    ), f"Lines found: {lines}\nErr output: {err}"
    assert "dev" not in info, f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize(
    "cmd,expected", (("--access-log", True), ("--no-access-log", False))
)
def test_access_logs(cmd: str, expected: bool):
    command = ["sanic", "fake.server.app", cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    info = read_app_info(lines)

    assert (
        info["access_log"] is expected
    ), f"Lines found: {lines}\nErr output: {err}"


@pytest.mark.parametrize("cmd", ("--version", "-v"))
def test_version(cmd: str):
    command = ["sanic", cmd]
    out, err, exitcode = capture(command)
    version_string = f"Sanic {__version__}; Routing {__routing_version__}\n"

    assert out == version_string.encode("utf-8")


@pytest.mark.parametrize(
    "cmd,expected",
    (
        ("--noisy-exceptions", True),
        ("--no-noisy-exceptions", False),
    ),
)
def test_noisy_exceptions(cmd: str, expected: bool):
    command = ["sanic", "fake.server.app", cmd]
    out, err, exitcode = capture(command)
    lines = out.split(b"\n")
    info = read_app_info(lines)

    assert (
        info["noisy_exceptions"] is expected
    ), f"Lines found: {lines}\nErr output: {err}"



import logging
import os
import platform
import sys

from unittest.mock import Mock

import pytest

from sanic import Sanic, __version__
from sanic.application.logo import BASE_LOGO
from sanic.application.motd import MOTD, MOTDTTY


def test_logo_base(app, run_startup):
    logs = run_startup(app)

    assert logs[0][1] == logging.DEBUG
    assert logs[0][2] == BASE_LOGO


def test_motd_with_expected_info(app, run_startup):
    logs = run_startup(app)

    assert logs[1][2] == f"Sanic v{__version__}"
    assert logs[3][2] == "mode: debug, single worker"
    assert logs[4][2] == "server: sanic, HTTP/1.1"
    assert logs[5][2] == f"python: {platform.python_version()}"
    assert logs[6][2] == f"platform: {platform.platform()}"


def test_motd_init():
    _orig = MOTDTTY.set_variables
    MOTDTTY.set_variables = Mock()
    motd = MOTDTTY(None, "", {}, {})

    motd.set_variables.assert_called_once()
    MOTDTTY.set_variables = _orig


def test_motd_display(caplog):
    motd = MOTDTTY("       foobar        ", "", {"one": "1"}, {"two": "2"})

    with caplog.at_level(logging.INFO):
        motd.display()

    version_line = f"Sanic v{__version__}".center(motd.centering_length)
    assert (
        "".join(caplog.messages)
        == f"""
  ┌────────────────────────────────┐
  │ {version_line} │
  │                                │
  ├───────────────────────┬────────┤
  │        foobar         │ one: 1 │
  |                       ├────────┤
  │                       │ two: 2 │
  └───────────────────────┴────────┘
"""
    )


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Not on 3.7")
def test_reload_dirs(app):
    app.config.LOGO = None
    app.config.AUTO_RELOAD = True
    app.prepare(reload_dir="./", auto_reload=True, motd_display={"foo": "bar"})

    existing = MOTD.output
    MOTD.output = Mock()

    app.motd("foo")

    MOTD.output.assert_called_once()
    assert (
        MOTD.output.call_args.args[2]["auto-reload"]
        == f"enabled, {os.getcwd()}"
    )
    assert MOTD.output.call_args.args[3] == {"foo": "bar"}

    MOTD.output = existing
    Sanic._app_registry = {}



"""
Sanic
"""
import codecs
import os
import re
import sys

from distutils.util import strtobool

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """
    Provide a Test runner to be used from setup.py to run unit tests
    """

    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


def open_local(paths, mode="r", encoding="utf8"):
    path = os.path.join(os.path.abspath(os.path.dirname(__file__)), *paths)

    return codecs.open(path, mode, encoding)


with open_local(["sanic", "__version__.py"], encoding="latin1") as fp:
    try:
        version = re.findall(
            r"^__version__ = \"([^']+)\"\r?$", fp.read(), re.M
        )[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

with open_local(["README.rst"]) as rm:
    long_description = rm.read()

setup_kwargs = {
    "name": "sanic",
    "version": version,
    "url": "http://github.com/sanic-org/sanic/",
    "license": "MIT",
    "author": "Sanic Community",
    "author_email": "admhpkns@gmail.com",
    "description": (
        "A web server and web framework that's written to go fast. "
        "Build fast. Run fast."
    ),
    "long_description": long_description,
    "packages": find_packages(),
    "package_data": {"sanic": ["py.typed"]},
    "platforms": "any",
    "python_requires": ">=3.7",
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    "entry_points": {"console_scripts": ["sanic = sanic.__main__:main"]},
}

env_dependency = (
    '; sys_platform != "win32" ' 'and implementation_name == "cpython"'
)
ujson = "ujson>=1.35" + env_dependency
uvloop = "uvloop>=0.5.3" + env_dependency
types_ujson = "types-ujson" + env_dependency
requirements = [
    "sanic-routing>=22.3.0,<22.6.0",
    "httptools>=0.0.10",
    uvloop,
    ujson,
    "aiofiles>=0.6.0",
    "websockets>=10.0",
    "multidict>=5.0,<7.0",
]

tests_require = [
    "sanic-testing>=22.3.0",
    "pytest==6.2.5",
    "coverage==5.3",
    "gunicorn==20.0.4",
    "pytest-cov",
    "beautifulsoup4",
    "pytest-sanic",
    "pytest-sugar",
    "pytest-benchmark",
    "chardet==3.*",
    "flake8",
    "black",
    "isort>=5.0.0",
    "bandit",
    "mypy>=0.901,<0.910",
    "docutils",
    "pygments",
    "uvicorn<0.15.0",
    "slotscheck>=0.8.0,<1",
    types_ujson,
]

docs_require = [
    "sphinx>=2.1.2",
    "sphinx_rtd_theme>=0.4.3",
    "docutils",
    "pygments",
    "m2r2",
    "enum-tools[sphinx]",
    "mistune<2.0.0",
]

dev_require = tests_require + [
    "cryptography",
    "tox",
    "towncrier",
]

all_require = list(set(dev_require + docs_require))

if strtobool(os.environ.get("SANIC_NO_UJSON", "no")):
    print("Installing without uJSON")
    requirements.remove(ujson)
    tests_require.remove(types_ujson)

# 'nt' means windows OS
if strtobool(os.environ.get("SANIC_NO_UVLOOP", "no")):
    print("Installing without uvLoop")
    requirements.remove(uvloop)

extras_require = {
    "test": tests_require,
    "dev": dev_require,
    "docs": docs_require,
    "all": all_require,
    "ext": ["sanic-ext"],
    "http3": ["aioquic"],
}

setup_kwargs["install_requires"] = requirements
setup_kwargs["tests_require"] = tests_require
setup_kwargs["extras_require"] = extras_require
setup_kwargs["cmdclass"] = {"test": PyTest}
setup(**setup_kwargs)



from enum import Enum, IntEnum, auto


class StrEnum(str, Enum):
    def _generate_next_value_(name: str, *args) -> str:  # type: ignore
        return name.lower()


class Server(StrEnum):
    SANIC = auto()
    ASGI = auto()
    GUNICORN = auto()


class Mode(StrEnum):
    PRODUCTION = auto()
    DEBUG = auto()


class ServerStage(IntEnum):
    STOPPED = auto()
    PARTIAL = auto()
    SERVING = auto()
